"""Chain that interprets a prompt and executes bash code to perform bash operations."""

from langchain.llms.base import BaseLLM
from langchain.prompts.base import BasePromptTemplate
from pydantic import BaseModel

from zamm.actions.use_terminal.chain import TerminalChain

from .bash_action_prompt import BASH_COMMAND_TEMPLATE


class ZBashChain(TerminalChain, BaseModel):
    """Custom Bash chain for easier experimentation"""

    llm: BaseLLM
    """LLM wrapper to use."""
    prompt: BasePromptTemplate = BASH_COMMAND_TEMPLATE



"""Agent state data."""

from abc import ABC, abstractmethod
from typing import Dict, List, Sequence, Tuple

from langchain.schema import AgentAction
from pydantic import BaseModel

from .step import StepOutput


class BaseAgentMemory(BaseModel, ABC):
    """Agent working memory, modeled off of Chain memory.

    Should contain everything needed for the agent to make its next decision. After
    all, how can what you don't remember possibly inform your decisions?
    """

    inputs: Dict[str, str]

    @abstractmethod
    def add_step(self, output: StepOutput) -> None:
        """Add step output to memory."""

    @abstractmethod
    def steps(self) -> Sequence[StepOutput]:
        """Return all steps so far.

        Uses Sequence instead of List so that this could return subclasses of
        StepOutput. This is because List is invariant but Sequence is covariant:

        https://mypy.readthedocs.io/en/stable/common_issues.html#variance
        """

    @abstractmethod
    def as_intermediate_steps(self) -> List[Tuple[AgentAction, str]]:
        """Return intermediate steps in unstructured form.

        For backwards compatibility.
        """


class AgentMemory(BaseAgentMemory):
    """Default Agent memory that assumes StepOutput type for all steps."""

    _steps: List[StepOutput] = []

    def add_step(self, output: StepOutput) -> None:
        """Add step output to memory."""
        self._steps.append(output)

    def steps(self) -> Sequence[StepOutput]:
        """Return all steps so far."""
        return self._steps

    def as_intermediate_steps(self) -> List[Tuple[AgentAction, str]]:
        """Return intermediate steps in unstructured form.

        For backwards compatibility.
        """
        return [output.as_intermediate_step() for output in self._steps]



from langchain.prompts.prompt import PromptTemplate

NOTE_PROMPT = PromptTemplate(
    input_variables=[],
    template="You note that: ",
)

NOTE_LOGGER = PromptTemplate(
    input_variables=["note"],
    template="\nYou note that {note}",
)



from typing import Any, List

from langchain.prompts.prompt import PromptTemplate

from zamm.prompts import PrefixedPromptTemplate
from zamm.utils import f_join


class TerminalPromptTemplate(PrefixedPromptTemplate):
    terminal_prompt: str = """
You proceed to use the terminal:

```bash
$ """.lstrip()

    @property
    def _prompt_type(self) -> str:
        """Return the prompt type key."""
        return "terminal"

    def format(self, **kwargs: Any) -> str:
        prefix_text = self.prefix.format(**kwargs)
        return f_join(
            "",
            [
                prefix_text,
                self.terminal_prompt,
            ],
        )


TERMINAL_USAGE_PREFIX = """
You proceed to use the terminal:

```bash"""

TERMINAL_USAGE_SUFFIX = "```"


class TerminalUsageLogger(PromptTemplate):
    input_variables: List[str] = ["command", "output"]
    template: str = """
$ {command}
{output}""".lstrip()



import os

# choose something that is unlikely to appear in code files
FILE_SEPARATOR = "~~~~~~~~~~"


class FilesTool:
    """Sets up file reading and writing"""

    def read(self, path: str) -> str:
        if not os.path.isfile(path):
            return f"Unfortunately, the file at `{path}` does not exist."

        with open(path) as f:
            contents = f.read()
            return f"""
The contents of the file at `{path}` are

{FILE_SEPARATOR}
{contents}
{FILE_SEPARATOR}
""".strip()

    def write(self, path: str, contents: str) -> str:
        try:
            with open(path, "w") as f:
                f.write(contents)
                return f"The file at `{path}` was successfully written to."
        except Exception as e:
            return f"The file at `{path}` could not be written to: {e}"



from typing import Any, Dict

from langchain.llms.base import BaseLLM
from langchain.schema import AgentAction

from zamm.actions.base import Action
from zamm.agents.z_step import ZStepOutput
from zamm.prompts.chained import ChainedPromptTemplate
from zamm.prompts.prefixed import Prefix

from .chain import NoteChain
from .prompt import NOTE_LOGGER, NOTE_PROMPT


class NoteOutput(ZStepOutput):
    @classmethod
    def from_chain_output(cls, output: Dict[str, Any]):
        return cls(
            decision=AgentAction(
                tool=output["action"],
                tool_input=output["note"],
                log="dummy log",
            ),
            observation="dummy observation",
            logger_template=NOTE_LOGGER,
        )

    @property
    def note(self):
        return self.decision.tool_input

    @property
    def template_args(self) -> Dict[str, str]:
        """Construct the dict used to render this output"""
        return {"note": self.note}


class MakeNote(Action):
    @classmethod
    def default(cls, llm: BaseLLM, prefix: Prefix):
        return cls(
            name="Make a mental note to yourself",
            output_type=NoteOutput,
            chain=NoteChain(
                llm=llm,
                prompt=ChainedPromptTemplate("", prefix, NOTE_PROMPT),
            ),
        )



from typing import Any, Dict, List

from langchain.chains.base import Chain


class ErrorChain(Chain):
    error_message: str

    @property
    def input_keys(self) -> List[str]:
        """Input keys this chain expects."""
        return []

    @property
    def output_keys(self) -> List[str]:
        """Output keys this chain expects."""
        return []

    def _call(self, inputs: Dict[str, Any]) -> Dict[str, str]:
        raise RuntimeError(self.error_message)



from .action_chain import ActionChain
from .choice.base import ChoiceChain
from .choice.prompt import ChoicePrompt, ChoicePromptTemplate
from .error import ErrorChain
from .no_op import NoOpChain

__all__ = [
    "ActionChain",
    "ChoiceChain",
    "ChoicePrompt",
    "ChoicePromptTemplate",
    "NoOpChain",
    "ErrorChain",
]



from langchain.prompts.prompt import PromptTemplate

from zamm.actions.use_terminal.prompt import TerminalPromptTemplate

MANAGER_PROMPT = "You are a manager who decides to give his subordinate the task: "
MANAGER_TEMPLATE = PromptTemplate(input_variables=[], template=MANAGER_PROMPT)


EMPLOYEE_DOCUMENTATION_INTRO = """
You are a button presser who has access to a Bash terminal. You have diligently pored over your employee training manual, which reads:

-----
{documentation}
-----

Now your boss has a task for you:

{task}""".lstrip()  # noqa

EMPLOYEE_DOCUMENTATION_PROMPT = PromptTemplate(
    input_variables=["documentation", "task"],
    template=EMPLOYEE_DOCUMENTATION_INTRO,
)

EMPLOYEE_TEACHING_INTRO = """
You are a state-of-the-art LLM, hired as an AI employee for the ZAMM firm. Your boss has asked you to perform the following task:

> {task}
{agent_scratchpad}
""".lstrip()  # noqa

EMPLOYEE_TEACHING_INTRO_TEMPLATE = PromptTemplate(
    input_variables=["task", "agent_scratchpad"],
    template=EMPLOYEE_TEACHING_INTRO,
)

FOLLOW_INSTRUCTIONS_INTRO = """
You are a simple button presser who simply follows instructions without doing things very creatively. You always follow every instruction, in order, until the task is done. This includes following instructions in the **Confirmation** section of your employee training manual.

You have access to a Bash terminal and a file editor. The Bash terminal is unable to edit files, so instead you always use the file editor for that.

You have diligently pored over your employee training manual, which reads:

-----
{documentation}
-----

Your boss has asked you to perform the following task:

> {task}

Fortunately, this is exactly the task that the training manual has prepared you for! You follow its instructions closely.

{agent_scratchpad}
""".lstrip()  # noqa

FOLLOW_INSTRUCTIONS_TEMPLATE = PromptTemplate(
    input_variables=["task", "documentation", "agent_scratchpad"],
    template=FOLLOW_INSTRUCTIONS_INTRO,
)

BASH_COMMAND_TEMPLATE = TerminalPromptTemplate(prefix=EMPLOYEE_DOCUMENTATION_PROMPT)


SUCCESS_CRITERIA_PROMPT = (
    EMPLOYEE_DOCUMENTATION_INTRO
    + """

You jot down a one-line version of each item in the success checklist for this task on a notepad:

1. """  # noqa
)

SUCCESS_CRITERIA_TEMPLATE = PromptTemplate(
    input_variables=["documentation", "task"], template=SUCCESS_CRITERIA_PROMPT
)



import langchain_visualizer  # isort:skip  # noqa: F401
import asyncio
import glob
import os
import sys
from enum import Enum
from importlib import resources
from typing import Callable, List, Optional

import typer
import vcr_langchain as vcr
import yaml
from appdirs import AppDirs
from langchain.llms.base import BaseLLM
from langchain.llms.openai import OpenAI
from ulid import ULID
from vcr.record_mode import RecordMode

from zamm.agents.employee import ZammEmployee
from zamm.chains.ask_task import AskForTaskChain
from zamm.llms.human import Human
from zamm.utils import current_directory

DOCUMENTATION_PATH = "documentation.zamm.md"
VISUALIZE = True
app_dirs = AppDirs(appname="zamm")
ZAMM_SESSION_PATH = app_dirs.user_data_dir + "/sessions"
INTERNAL_TUTORIAL_PREFIX = "@internal"
INTERNAL_TUTORIAL_PACKAGE = "zamm.resources.tutorials"


app = typer.Typer(pretty_exceptions_show_locals=False)


def visualize_or_run(fn: Callable):
    # todo: change ICE to allow for running from other modules
    fn.__module__ = "__main__"

    if VISUALIZE:
        # ICE has its own defopts
        sys.argv = sys.argv[:1]
        return langchain_visualizer.visualize(fn)
    else:
        return asyncio.get_event_loop().run_until_complete(fn())


def run_chain(cassette_path: str, run: Callable):
    with vcr.use_cassette(
        path=cassette_path,
        record_mode=RecordMode.NEW_EPISODES,
    ):
        with current_directory(os.getcwd()):
            try:
                visualize_or_run(run)
            except RuntimeError:
                print(f"Exiting, session recording should be saved to {cassette_path}")


def play_interactions(
    llm: BaseLLM,
    cassette_path: str,
    tutorial_output_path: str,
):
    async def run():
        task = AskForTaskChain(llm=llm)({})
        # always trust shell commands from the person who invoked this program in the
        # first place
        employee = ZammEmployee(llm=llm, terminal_safe_mode=False)
        result = employee(task)
        if result is not None:
            with open(tutorial_output_path, "w") as f:
                try:
                    f.write(result["output"])
                    print(f"Tutorial saved to {tutorial_output_path}")
                except Exception as e:
                    print(f"Could not save tutorial to {tutorial_output_path}: {e}")

    run_chain(cassette_path, run)


def execute_llm_task(
    employee: ZammEmployee,
    task: str,
    tutorial: Optional[str],
    cassette_path: str,
):
    async def run():
        args = {"task": task}
        if tutorial is not None:
            args["documentation"] = tutorial
        result = employee(args)
        if result is not None:
            print(
                "LLM indicates it has completed the task. Session recording should be "
                f"saved to {cassette_path}"
            )

    run_chain(cassette_path, run)


SESSION_RECORD_OPTION = typer.Option(
    None,
    help="Recorded interactions from a previous unfinished session.",
)

OUTPUT_OPTION = typer.Option(
    None,
    help="File to output the tutorial to.",
)


def get_cassette_path(cassette: Optional[typer.FileText]):
    if cassette is None:
        os.makedirs(ZAMM_SESSION_PATH, exist_ok=True)
        return f"{ZAMM_SESSION_PATH}/session_{ULID()}.yaml"
    else:
        cassette.close()
        return cassette.name


def get_cassette_and_output(
    cassette: Optional[typer.FileText],
    output: Optional[typer.FileTextWrite],
):
    cassette_path = get_cassette_path(cassette)

    if output is None:
        session_prefix = os.path.splitext(cassette_path)[0]
        output_path = f"{session_prefix}.md"
    else:
        output_path = output.name
        output.close()

    return cassette_path, output_path


def get_last_session():
    # https://stackoverflow.com/a/39327156
    files = glob.glob(f"{ZAMM_SESSION_PATH}/session_*.yaml")
    if len(files) == 0:
        print("No sessions found.")
        sys.exit(1)
    return max(files, key=os.path.getctime)


@app.command()
def teach(
    session_recording: Optional[typer.FileText] = SESSION_RECORD_OPTION,
    last_session: bool = typer.Option(
        False,
        help="The last session that was in progress",
    ),
    output: Optional[typer.FileTextWrite] = OUTPUT_OPTION,
):
    """Record a tutorial interaction."""

    cassette_path, output_path = get_cassette_and_output(
        cassette=session_recording, output=output
    )

    if last_session:
        cassette_path = get_last_session()

    llm = Human()
    play_interactions(
        llm=llm, cassette_path=cassette_path, tutorial_output_path=output_path
    )


@app.command()
def re_record(
    session_recording: Optional[typer.FileText] = typer.Option(
        None,
        help="Recorded interactions from a previous unfinished session.",
    ),
    last_session: bool = typer.Option(
        False,
        help="The last session that was in progress",
    ),
    output: Optional[typer.FileTextWrite] = OUTPUT_OPTION,
):
    """Re-record a tutorial interaction.

    Keep all inputs the same. Useful for when you're making cosmetic changes to the
    prompting, but wish to otherwise keep everything the same.
    """
    if not session_recording and not last_session:
        print("You must either specify session-recording or use --last-session.")
        sys.exit(1)

    cassette_path, output_path = get_cassette_and_output(
        cassette=session_recording, output=output
    )

    if last_session:
        cassette_path = get_last_session()

    try:
        with open(cassette_path) as c:
            interactions = yaml.load(c, Loader=yaml.Loader)["interactions"]
        inputs = [
            i["response"]
            for i in interactions
            if i["request"]["uri"].startswith("tool://Human")
        ]
        llm = Human(prerecorded_responses=inputs)
        os.remove(cassette_path)
        play_interactions(llm, cassette_path, tutorial_output_path=output_path)
    except yaml.YAMLError as exc:
        print(exc)


class Safety(str, Enum):
    # todo: use Python's Flag instead
    on = "on"
    off = "off"


@app.command()
def execute(
    task: str = typer.Option(
        ...,
        help="What you'd like the LLM to do",
    ),
    documentation: Optional[str] = typer.Option(
        None,
        help=(
            "Documentation file to help the LLM accomplish the task. Prefix with "
            "@internal for internal help files."
        ),
    ),
    session_recording: Optional[typer.FileText] = SESSION_RECORD_OPTION,
    model: str = typer.Option(
        "text-davinci-003",
        help="What OpenAI large language model to use for execution",
    ),
    safety: Safety = typer.Option(
        Safety.on,
        help=(
            "If on, will ask user to confirm every terminal command. If off, will run "
            "LLM commands automatically, WHICH MAY EXPOSE YOU TO LLM ATTACKS."
        ),
    ),
):
    """Ask the LLM to do something."""

    cassette_path = get_cassette_path(cassette=session_recording)

    llm = OpenAI(model_name=model, temperature=0, max_tokens=-1)
    if session_recording is not None:
        session_recording.close()
    if documentation is None:
        tutorial = None
    else:
        if documentation.startswith(INTERNAL_TUTORIAL_PREFIX):
            internal_path = documentation[len(INTERNAL_TUTORIAL_PREFIX) + 1 :]
            if not internal_path.endswith(".md"):
                internal_path += ".md"
            tutorial = resources.read_text(INTERNAL_TUTORIAL_PACKAGE, internal_path)
        else:
            with open(documentation) as f:
                tutorial = f.read()
    employee = ZammEmployee(llm=llm, terminal_safe_mode=safety.value == "on")
    execute_llm_task(
        employee=employee, task=task, tutorial=tutorial, cassette_path=cassette_path
    )


@app.command()
def prompt(
    raw: Optional[typer.FileText] = typer.Option(
        ...,
        help="Raw prompt to be sent to the LLM verbatim",
    ),
    stop: List[str] = typer.Option(
        [],
        help="LLM stopping tokens",
    ),
):
    llm = OpenAI(temperature=0, max_tokens=-1)
    escaped_stops = [s.encode("utf-8").decode("unicode_escape") for s in stop]

    async def run():
        result = llm(prompt=raw.read(), stop=escaped_stops)
        print(result)

    visualize_or_run(run)



"""Chain that chooses and performs the next action."""
from abc import ABC, abstractmethod
from typing import Any, Dict, List

from langchain.chains.base import Chain
from pydantic import BaseModel

from zamm.utils import safe_call


class OptionsChain(Chain, BaseModel, ABC):
    option_picker: Chain
    options: List[Chain]

    @property
    def input_keys(self) -> List[str]:
        """Input keys to this chain."""
        return self.option_picker.input_keys

    @property
    def output_keys(self) -> List[str]:
        """Possible output keys produced by this chain."""
        all_keys = set(self.option_picker.output_keys)
        for option in self.options:
            all_keys.update(option.output_keys)
        return list(all_keys)

    @abstractmethod
    def option_index(self, picker_output: Dict[str, Any]) -> int:
        """The index of the option chosen."""

    def _validate_outputs(self, outputs: Dict[str, str]) -> None:
        """Always validated because different options may produce different outputs."""

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        """Run the logic of this chain and return the output."""
        picker_output = self.option_picker(inputs)
        option_index = self.option_index(picker_output)
        option = self.options[option_index]
        full_inputs = {**inputs, **picker_output}
        option_output = safe_call(option, full_inputs)
        return {**full_inputs, **option_output}



"""Chain that interprets a prompt and executes bash code to perform bash operations."""
from typing import Dict, List

from langchain.chains.base import Chain
from langchain.llms.base import BaseLLM
from langchain.prompts.base import BasePromptTemplate
from pydantic import BaseModel

from .bash_action_prompt import MANAGER_TEMPLATE


class AskForTaskChain(Chain, BaseModel):
    """Ask for a task to perform"""

    llm: BaseLLM
    """LLM wrapper to use."""
    prompt: BasePromptTemplate = MANAGER_TEMPLATE
    output_key: str = "task"

    @property
    def input_keys(self) -> List[str]:
        """Input keys to this chain."""
        return []

    @property
    def output_keys(self) -> List[str]:
        """Output keys produced by this chain."""
        return [self.output_key]

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        output = self.llm(self.prompt.format(**inputs), stop=["\n"]).strip()

        return {self.output_key: output}



from typing import Dict, List

from langchain.chains.llm import LLMChain
from pydantic import BaseModel

from .terminal import ZTerminal


class TerminalChain(LLMChain, BaseModel):
    """Asks LLM for terminal input, and executes it"""

    terminal: ZTerminal

    @property
    def input_keys(self) -> List[str]:
        """Input keys to this chain."""
        return self.prompt.input_variables

    @property
    def output_keys(self) -> List[str]:
        return ["command", "output"]

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        command = self.llm(self.prompt.format(**inputs), stop=["\n"]).strip()
        output = self.terminal.run_bash_command(command).rstrip()

        return {"command": command, "output": output}



from typing import Any, Dict, Type

from langchain.chains.base import Chain
from langchain.llms.base import BaseLLM
from langchain.tools.base import BaseTool

from zamm.agents.z_step import DummyStepOutput, ZStepOutput
from zamm.chains.dummy import DummyLLMChain


def dummy_func(input: str) -> str:
    raise NotImplementedError()


class Action(BaseTool):
    name: str
    description: str = "Dummy description"
    chain: Chain
    output_type: Type[ZStepOutput]

    def _run(self, tool_input: str) -> str:
        raise NotImplementedError()

    async def _arun(self, tool_input: str) -> str:
        raise NotImplementedError()

    @property
    def choice_text(self) -> str:
        return self.name

    def use(self, inputs: Dict[str, str]) -> Dict[str, Any]:
        return self.chain(inputs)

    def construct_output(self, output: Dict[str, Any]) -> ZStepOutput:
        """Create a structured representation of tool output."""
        return self.output_type.from_chain_output(output)


class DummyAction(Action):
    def __init__(
        self, llm: BaseLLM, output_type: Type[ZStepOutput] = DummyStepOutput, **kwargs
    ):
        super().__init__(
            chain=DummyLLMChain(llm=llm), output_type=output_type, **kwargs
        )
        assert self.chain is not None



import re
from typing import Dict, List

from langchain.chains.llm import LLMChain
from pydantic import BaseModel

from .prompt import ChoicePromptTemplate


class ChoiceChain(LLMChain, BaseModel):
    """Returns LLM selection of a variety of options"""

    prompt: ChoicePromptTemplate
    choice_num_key: str = "choice_num"
    choice_key: str = "choice"

    @property
    def input_keys(self) -> List[str]:
        """Input keys to this chain."""
        return self.prompt.input_variables

    @property
    def output_keys(self) -> List[str]:
        return [self.choice_num_key, self.choice_key]

    def _return_dict(self, choice_num: int, choice: str) -> Dict[str, str]:
        return {self.choice_num_key: str(choice_num), self.choice_key: choice}

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        prompt = self.prompt.format(**inputs)
        result = self.llm(prompt, stop=[" "])

        regex_search = re.search(r"\d+", result)
        assert regex_search is not None, f"No number in response '{result}'"
        choice_num = int(regex_search.group())
        if not 1 <= choice_num <= len(self.prompt.choices):
            raise Exception(f"Unknown choice: '{result}'")
        choice = self.prompt.choices[choice_num - 1]

        return self._return_dict(choice_num, choice)



from typing import Any, List, Optional, Union

from langchain.prompts import BasePromptTemplate, PromptTemplate

Prefix = Union[str, BasePromptTemplate]


class PrefixedPromptTemplate(BasePromptTemplate):
    prefix: BasePromptTemplate

    def __init__(
        self,
        prefix: Prefix,
        input_variables: Optional[List[str]] = None,
        **kwargs: Any,
    ):
        if isinstance(prefix, str):
            prefix_input_vars = input_variables or []
            prefix = PromptTemplate(input_variables=prefix_input_vars, template=prefix)
        elif input_variables is not None:
            raise ValueError("Cannot both prompt template and custom input_variables")
        super().__init__(
            prefix=prefix,
            input_variables=prefix.input_variables,
            **kwargs,
        )



from langchain.agents.tools import Tool

from zamm.agents.tools.files import FilesTool


def _get_files_read() -> Tool:
    return Tool(
        "File Reader",
        FilesTool().read,
        (
            "Use this to read files. Input should be the file path. Output will be a "
            "description of the file, including the file contents."
        ),
    )


def _get_files_write() -> Tool:
    return Tool(
        "File Writer",
        FilesTool().write,  # type: ignore
        (
            "Use this to write files. Inputs should be the file path and the content "
            "to be written to the file. Output will be a description of the results "
            "of the write."
        ),
    )



from langchain.prompts.prompt import PromptTemplate

WHICH_FILE_PROMPT = PromptTemplate(
    input_variables=[],
    template="You decide to edit the file located at: `",
)

NEW_CONTENTS_PROMPT = PromptTemplate(
    input_variables=["file_path"],
    template="""
You decide to edit the file `{file_path}`. It currently does not exist.

You write this content out to the file:

```
""".lstrip(),
)

REPLACE_CONTENTS_PROMPT = PromptTemplate(
    input_variables=["file_path", "old_contents"],
    template="""
You decide to edit the file `{file_path}`. Its current contents are

```
{old_contents}
```

You replace the file contents with

```
""".lstrip(),
)

NEW_FILE_LOGGER = PromptTemplate(
    input_variables=["file_path", "new_contents"],
    template="""
You decide to edit the file `{file_path}`. It doesn't yet exist.

You write out to the file the contents

```
{new_contents}
```
""".rstrip(),
)

EDIT_LOGGER = PromptTemplate(
    input_variables=["file_path", "old_contents", "new_contents"],
    template="""
You decide to edit the file `{file_path}`. Its old contents were

```
{old_contents}
```

You replace the file contents with

```
{new_contents}
```
""".rstrip(),
)



from typing import Dict, List

from langchain.chains.llm import LLMChain
from pydantic import BaseModel


class NoteChain(LLMChain, BaseModel):
    """Asks LLM for what it notes"""

    @property
    def input_keys(self) -> List[str]:
        """Input keys to this chain."""
        return self.prompt.input_variables

    @property
    def output_keys(self) -> List[str]:
        return ["note"]

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        note = self.llm(self.prompt.format(**inputs), stop=["\n"]).strip()
        return {"note": note}



from typing import Any, Dict, Optional

from langchain.llms.base import BaseLLM
from langchain.schema import AgentAction

from zamm.actions.base import Action
from zamm.agents.z_step import StepOutput, ZStepOutput
from zamm.prompts.prefixed import Prefix
from zamm.utils import safe_format

from .chain import EditFileChain
from .prompt import EDIT_LOGGER, NEW_FILE_LOGGER


class EditFileOutput(ZStepOutput):
    file_exists: bool
    old_contents: Optional[str]

    @classmethod
    def from_chain_output(cls, output: Dict[str, Any]):
        file_exists = output["file_exists"]
        return cls(
            decision=AgentAction(
                tool=output["action"],
                tool_input=output["file_path"],
                log="dummy log",
            ),
            observation=output["new_contents"],
            old_contents=output["old_contents"] if file_exists else None,
            file_exists=file_exists,
            logger_template=EDIT_LOGGER,
        )

    @property
    def file_path(self):
        return self.decision.tool_input

    @property
    def new_contents(self):
        return self.observation

    @property
    def template_args(self) -> Dict[str, str]:
        """Construct the dict used to render this output"""
        args = {
            "file_path": self.file_path,
            "file_exists": self.file_exists,
            "new_contents": self.new_contents,
        }
        if self.file_exists:
            assert self.old_contents is not None
            args["old_contents"] = self.old_contents
        return args

    def _log(self, previous: StepOutput | None, next: StepOutput | None) -> str:
        if self.file_exists:
            return safe_format(EDIT_LOGGER, self.template_args)
        else:
            return safe_format(NEW_FILE_LOGGER, self.template_args)


class EditFile(Action):
    @classmethod
    def default(cls, llm: BaseLLM, prefix: Prefix):
        return cls(
            name="Edit a file",
            output_type=EditFileOutput,
            chain=EditFileChain.default(
                llm=llm,
                prefix=prefix,
            ),
        )



"""Chain that chooses and performs the next action."""
from typing import Any, Dict, List, Union

from langchain.chains.base import Chain
from pydantic import BaseModel

from zamm.actions.base import Action

from .choice.base import ChoiceChain
from .options_chain import OptionsChain


class DummyChain(Chain, BaseModel):
    @property
    def input_keys(self) -> List[str]:
        raise NotImplementedError()

    @property
    def output_keys(self) -> List[str]:
        return []

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        raise NotImplementedError()


class ActionChain(OptionsChain, BaseModel):
    option_picker: ChoiceChain
    actions: List[Action]
    step_output_key: str = "step_output"

    def __init__(self, actions: List[Action], **kwargs):
        options = [action.chain for action in actions]
        super().__init__(actions=actions, options=options, **kwargs)

    @property
    def output_keys(self) -> List[str]:
        """Possible output keys produced by this chain."""
        return super().output_keys + [self.step_output_key]

    @property
    def choice_num_key(self) -> str:
        return self.option_picker.choice_num_key

    def option_index(self, picker_output: Dict[str, Any]) -> int:
        """The index of the option chosen."""
        return int(picker_output[self.choice_num_key]) - 1

    def __call__(
        self,
        inputs: Union[Dict[str, Any], Any],
        return_only_outputs: bool = False,
    ) -> Dict[str, Any]:
        output = super().__call__(inputs, return_only_outputs)
        action = self.actions[self.option_index(output)]
        output[self.step_output_key] = action.output_type.from_chain_output(output)
        return output



"""Chain that chooses and performs the next action."""
from abc import abstractmethod
from typing import Any, Dict, List, Optional, Union

from langchain.chains.base import Chain

from .options_chain import OptionsChain


class BooleanSwitchChain(OptionsChain):
    boolean_output_key: Optional[str]

    def __init__(self, true_chain: Chain, false_chain: Chain, **kwargs):
        super().__init__(options=[false_chain, true_chain], **kwargs)

    @property
    def output_keys(self) -> List[str]:
        """Possible output keys produced by this chain."""
        keys = super().output_keys
        if self.boolean_output_key is not None:
            keys.append(self.boolean_output_key)
        return keys

    @abstractmethod
    def condition_was_met(self, outputs: Dict[str, Any]) -> bool:
        """Whether or not the condition was met."""

    def option_index(self, picker_output: Dict[str, Any]) -> int:
        """The index of the option chosen."""
        if self.condition_was_met(picker_output):
            return 1
        return 0

    def __call__(
        self,
        inputs: Union[Dict[str, Any], Any],
        return_only_outputs: bool = False,
    ) -> Dict[str, Any]:
        output = super().__call__(inputs, return_only_outputs)
        if self.boolean_output_key is not None:
            output[self.boolean_output_key] = self.condition_was_met(output)
        return output



from typing import Dict, Optional, Sequence, Tuple

from langchain.agents.agent import Agent
from langchain.llms.base import BaseLLM
from langchain.prompts.base import BasePromptTemplate
from langchain.tools import BaseTool
from pydantic import root_validator

from zamm.chains.dummy import DummyLLMChain


class CustomAgent(Agent):
    """Custom agent that sidesteps a lot of langchain's opinions about how agents
    should work."""

    def _extract_tool_and_input(self, text: str) -> Optional[Tuple[str, str]]:
        """Extract tool and tool input from llm output."""
        raise NotImplementedError()

    @property
    def observation_prefix(self) -> str:
        """Prefix to append the observation with."""
        raise NotImplementedError()

    @property
    def llm_prefix(self) -> str:
        """Prefix to append the LLM call with."""
        raise NotImplementedError()

    @classmethod
    def create_prompt(cls, tools: Sequence[BaseTool]) -> BasePromptTemplate:
        """Create a prompt for this class."""
        raise NotImplementedError()

    @root_validator()
    def validate_prompt(cls, values: Dict) -> Dict:
        """Ignore parent prompt validation"""
        return values

    @classmethod
    def from_llm(cls, llm: BaseLLM):
        return cls(llm_chain=DummyLLMChain(llm=llm))


class DummyAgent(CustomAgent):
    """Used when you want to use langchain's AgentExecutor but not Agent"""

    @property
    def _agent_type(self) -> str:
        """Return Identifier of agent type."""
        return "zamm-dummy"



import os
import re
from contextlib import contextmanager
from typing import Any

from fvalues import F
from langchain.chains.base import Chain
from langchain.prompts import BasePromptTemplate


def safe_inputs(
    lc_object: Chain | BasePromptTemplate, inputs: dict[str, str]
) -> dict[str, str]:
    if isinstance(lc_object, Chain):
        input_variables = lc_object.input_keys
    elif isinstance(lc_object, BasePromptTemplate):
        input_variables = lc_object.input_variables
    else:
        raise ValueError(f"Unknown type for {lc_object}")
    return {k: v for k, v in inputs.items() if k in input_variables}


def safe_call(chain: Chain, inputs: dict[str, str]) -> dict[str, Any]:
    return chain(safe_inputs(chain, inputs))


def safe_format(template: BasePromptTemplate, inputs: dict[str, str]) -> str:
    return template.format(**safe_inputs(template, inputs))


def f_join(joiner: str, substrings: list[str | F]) -> F:
    if substrings == []:
        return F("")

    joined = ""
    parts = []
    for substring in substrings:
        joined += substring
        if isinstance(substring, str):
            parts.append(substring)
        else:
            parts.extend(list(substring.parts))

        # if it's the empty string, we can just avoid polluting parts with it
        if joiner != "":
            joined += joiner
            parts.append(joiner)

    if joiner != "":  # joiner should only exist in between parts
        joined = joined[: -len(joiner)]
        parts.pop()

    return F(joined, parts=tuple(parts))


def ansi_escape_regex():
    escape_command = r"\[([\d;]*m|\d[A-K])"
    rs = r"\r|\\r"
    escape_prefixes = "|".join([r"\\033", r"\\e", r"\x1b"])
    escapes = f"({escape_prefixes}){escape_command}"
    return re.compile("|".join([rs, escapes]))


def remove_ansi_escapes(input):
    """ANSI escape remover.

    Incomplete escape interpreter based on
    https://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html
    """
    regex = ansi_escape_regex()
    cleaned = []
    for line in input.split("\n"):
        next_match = re.search(regex, line)
        while next_match is not None:
            full_match = next_match.group()
            _, command = next_match.groups()
            if full_match == "\r" or full_match == "\\r" or command.endswith("K"):
                line = line[next_match.end() :]
            elif command.endswith("A"):
                num_lines = int(command[:-1])
                cleaned = cleaned[:-num_lines]
                line = line[next_match.end() :]
            else:  # just remove the ANSI escape code if we can't interpret it
                line = line[: next_match.start()] + line[next_match.end() :]
            next_match = re.search(regex, line)
        cleaned.append(line)
    return "\n".join(cleaned)


@contextmanager
def current_directory(path: str):
    og_cwd = os.getcwd()
    os.makedirs(path, exist_ok=True)
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(og_cwd)



import os
import re
from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional

from pydantic import BaseModel
from vcr.cassette import Cassette
from vcr_langchain.patch import GenericPatch, add_patchers


@dataclass
class FileRead:
    file_exists: bool
    contents: Optional[str] = None

    def as_llm_output(self, contents_key: str = "contents") -> Dict[str, Any]:
        output: Dict[str, Any] = {"file_exists": self.file_exists}
        if self.file_exists:
            output[contents_key] = self.contents
        else:
            output[contents_key] = None
        return output


class FileSystemTool(BaseModel):
    """Tool for file system interactions."""

    spaces_in_tab = 8
    """If there's this many spaces at the beginning of the line, it gets treated as a
    tab.
    """

    def interpret_path(self, file_path: str) -> str:
        """Makes sure that ~ gets interpreted as home directory instead of filename."""
        return os.path.expanduser(file_path)

    @property
    def tab_regex(self):
        return re.compile(f"^ {{{self.spaces_in_tab}}}")

    def _spaces_to_tabs_per_line(self, line: str) -> str:
        while re.search(self.tab_regex, line) is not None:
            line = re.sub(self.tab_regex, "\t", line)
        return line

    def spaces_to_tabs(self, input: str) -> str:
        lines = input.split("\n")
        tabbed_lines = [self._spaces_to_tabs_per_line(line) for line in lines]
        return "\n".join(tabbed_lines)

    def read_file(self, file_path: str) -> FileRead:
        file_path = self.interpret_path(file_path)
        if os.path.isfile(file_path):
            with open(file_path) as f:
                return FileRead(file_exists=True, contents=f.read())
        return FileRead(file_exists=False)

    def write_file(self, file_path: str, contents: str) -> bool:
        file_path = self.interpret_path(file_path)
        folder = os.path.dirname(file_path)
        if folder != "":
            os.makedirs(folder, exist_ok=True)
        with open(file_path, "w") as f:
            f.write(self.spaces_to_tabs(contents))
        return True


class FileReadPatch(GenericPatch):
    def __init__(self, cassette: Cassette):
        super().__init__(cassette, FileSystemTool, "read_file")

    def get_same_signature_override(self) -> Callable:
        def _call(og_self: FileSystemTool, file_path: str) -> FileRead:
            """Same signature override patched into FileSystemTool"""
            return self.generic_override(og_self, file_path=file_path)

        return _call


class FileWritePatch(GenericPatch):
    def __init__(self, cassette: Cassette):
        super().__init__(cassette, FileSystemTool, "write_file")

    def get_same_signature_override(self) -> Callable:
        def _call(og_self: FileSystemTool, file_path: str, contents: str) -> FileRead:
            """Same signature override patched into FileSystemTool"""
            return self.generic_override(
                og_self, file_path=file_path, contents=contents
            )

        return _call


add_patchers(FileReadPatch, FileWritePatch)



import os
import shlex
import time
from typing import Callable

import pexpect
from pydantic import BaseModel
from vcr.cassette import Cassette
from vcr_langchain.patch import GenericPatch, add_patchers

from zamm.chains.human_cmd_review import HumanCommandReviewChain
from zamm.llms.human import Human
from zamm.utils import remove_ansi_escapes

ALWAYS_EXECUTE = False


class UnknownResult(Exception):
    pass


class ZTerminal(BaseModel):
    """A virtual terminal that supports interactive shell commands."""

    shell: pexpect.spawn
    bash_prompt: str
    refresh_interval: float
    output_size: int
    safe_mode: bool = True

    class Config:
        arbitrary_types_allowed = True

    def __init__(
        self, refresh_interval: float = 0.1, output_size: int = 1000, **kwargs
    ):
        sh = pexpect.spawn("/bin/bash", encoding="utf-8")
        time.sleep(refresh_interval)
        bash_prompt = sh.read_nonblocking(size=output_size)
        super().__init__(
            refresh_interval=refresh_interval,
            output_size=output_size,
            shell=sh,
            bash_prompt=bash_prompt,
            **kwargs,
        )

    @property
    def prompt_length(self) -> int:
        return len(self.bash_prompt)

    def _get_raw_shell_update_uncached(self, cmd: str) -> str:
        self.shell.sendline(cmd)
        results = ""
        try:
            while not results.endswith(self.bash_prompt):
                latest_output = self.shell.read_nonblocking(size=self.output_size)
                results += latest_output
                time.sleep(0.1)
        except pexpect.TIMEOUT as e:
            raise UnknownResult(
                "Terminal output does not have initial prompt of: "
                f"'{self.bash_prompt}':\n\n{results}"
            ) from e

        # todo: more robust way of syncing terminal actions to action chain state
        parsed_cmd = shlex.split(cmd)
        if len(parsed_cmd) == 2 and parsed_cmd[0] == "cd":
            os.chdir(parsed_cmd[1])

        return results

    def _get_raw_shell_update(self, cmd: str) -> str:
        return self._get_raw_shell_update_uncached(cmd)

    def _is_terminal_state_command(self, cmd: str) -> bool:
        return cmd.startswith("cd ")

    def _get_shell_update(self, cmd: str) -> str:
        if ALWAYS_EXECUTE or self._is_terminal_state_command(cmd):
            return self._get_raw_shell_update_uncached(cmd)
        return self._get_raw_shell_update(cmd)

    def run_bash_command(self, cmd: str) -> str:
        if self.safe_mode:
            cmd = HumanCommandReviewChain(llm=Human())(cmd)["command"]
        results = self._get_shell_update(cmd)
        assert results.startswith(cmd), (
            f"'{results}' does not start with '{cmd}'. "
            "Is non-ASCII terminal input involved?"
        )
        unix_results = results.replace("\r\n", "\n")
        # + 1 to remove leading "\n" after command input
        output = unix_results[len(cmd) + 1 : -self.prompt_length]
        without_ansi = remove_ansi_escapes(output)
        return without_ansi


class ZTerminalPatch(GenericPatch):
    def __init__(self, cassette: Cassette):
        super().__init__(cassette, ZTerminal, "_get_raw_shell_update")

    def get_same_signature_override(self) -> Callable:
        def _call(og_self: ZTerminal, cmd: str) -> str:
            """Same signature override patched into Human"""
            return self.generic_override(og_self, cmd=cmd)

        return _call


add_patchers(ZTerminalPatch)



from langchain.agents.load_tools import _BASE_TOOLS

from .load_tools import _get_files_read, _get_files_write

_BASE_TOOLS["file-reader"] = _get_files_read
_BASE_TOOLS["file-writer"] = _get_files_write



import os
import warnings
from typing import Callable, List, Optional

from langchain.llms.base import LLM
from simple_term_menu import TerminalMenu
from vcr.cassette import Cassette
from vcr_langchain.patch import GenericPatch, add_patchers

from zamm.chains.general import ChoicePrompt


class Human(LLM):
    prerecorded_responses: Optional[List[str]] = None
    playback_index: int = 0

    @property
    def _llm_type(self) -> str:
        return "Human"

    def _stop_hit(self, input: str, stops: List[str]) -> Optional[str]:
        for stop in stops:
            if stop in input:
                return stop
        return None

    def _call(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        """Run the LLM on the given prompt and input."""
        if self.prerecorded_responses and self.playback_index < len(
            self.prerecorded_responses
        ):
            response = self.prerecorded_responses[self.playback_index]
            self.playback_index += 1
            total_responses = len(self.prerecorded_responses)
            print(
                f"Replaying response #{self.playback_index} of {total_responses}",
                end="\r",
            )
            return response

        os.system("clear")

        if isinstance(prompt, ChoicePrompt):
            print(prompt.prefix_text)
            print()
            result = TerminalMenu(prompt.choices).show()
            if result is None:
                raise RuntimeError()  # most likely because user hit ctrl-C
            return str(result + 1)

        user_input = input(prompt)

        if stop is None or stop == ["\n"]:
            return user_input

        stop_hit = self._stop_hit(user_input, stop)
        while stop_hit is None:
            new_input = input()
            stop_hit = self._stop_hit(new_input, stop)
            user_input += "\n" + new_input

        separated_inputs = [x for x in user_input.split(stop_hit) if x]
        result = separated_inputs[0]
        if len(separated_inputs) > 1:
            warnings.warn(f"Ignoring rest of input after stop: {separated_inputs[1:]}")

        return result


class HumanPatch(GenericPatch):
    def __init__(self, cassette: Cassette):
        super().__init__(cassette, Human, "_call")

    def get_same_signature_override(self) -> Callable:
        def _call(og_self: Human, prompt: str, stop: Optional[List[str]] = None) -> str:
            """Same signature override patched into Human"""
            return self.generic_override(og_self, prompt=prompt, stop=stop)

        return _call


add_patchers(HumanPatch)



from typing import Any, Dict

from langchain.chains.llm import LLMChain
from langchain.llms.base import BaseLLM

from zamm.prompts.dummy import DummyPromptTemplate


class DummyLLMChain(LLMChain):
    def __init__(self, llm: BaseLLM):
        super().__init__(llm=llm, prompt=DummyPromptTemplate())

    def _call(self, inputs: Dict[str, Any]) -> Dict[str, str]:
        raise NotImplementedError()



from abc import abstractmethod
from typing import Any, Dict

from langchain.prompts.base import BasePromptTemplate

from zamm.prompts.dummy import DummyPromptTemplate

from .step import StepOutput


class ZStepOutput(StepOutput):
    logger_template: BasePromptTemplate

    def log(self, **kwargs) -> str:
        return self._log(kwargs["previous"], kwargs["next"])

    def _log(self, previous: StepOutput | None, next: StepOutput | None) -> str:
        return self.logger_template.format(**self.template_args)

    @classmethod
    @abstractmethod
    def from_chain_output(cls, output: Dict[str, Any]):
        """Construct the step from chain output"""

    @property
    @abstractmethod
    def template_args(self) -> Dict[str, str]:
        """Construct the dict used to render this output"""


class DummyStepOutput(ZStepOutput):
    logger_template: BasePromptTemplate = DummyPromptTemplate()

    @classmethod
    def from_chain_output(cls, output: Dict[str, Any]):
        """Construct the step from chain output"""
        raise NotImplementedError()

    @property
    def template_args(self) -> Dict[str, str]:
        """Construct the dict used to render this output"""
        raise NotImplementedError()



from typing import Any, Dict

from langchain.llms.base import BaseLLM
from langchain.prompts.prompt import PromptTemplate
from langchain.schema import AgentAction

from zamm.actions.base import Action
from zamm.agents.step import StepOutput
from zamm.agents.z_step import ZStepOutput
from zamm.prompts.chained import ChainedPromptTemplate
from zamm.prompts.prefixed import Prefix
from zamm.utils import safe_format

from .chain import TerminalChain
from .prompt import (
    TERMINAL_USAGE_PREFIX,
    TERMINAL_USAGE_SUFFIX,
    TerminalPromptTemplate,
    TerminalUsageLogger,
)
from .terminal import ZTerminal


class TerminalOutput(ZStepOutput):
    @classmethod
    def from_chain_output(cls, output: Dict[str, Any]):
        return cls(
            decision=AgentAction(
                tool=output["action"],
                tool_input=output["command"],
                log="dummy log",
            ),
            observation=output["output"],
            logger_template=TerminalUsageLogger(),
        )

    def _log(self, previous: StepOutput | None, next: StepOutput | None) -> str:
        template = self.logger_template
        if self.output == "":
            template = PromptTemplate(
                input_variables=["command"], template="$ {command}"
            )
        if previous is None or not isinstance(previous, TerminalOutput):
            template = ChainedPromptTemplate("\n", TERMINAL_USAGE_PREFIX, template)
        if next is None or not isinstance(next, TerminalOutput):
            template = ChainedPromptTemplate("\n", template, TERMINAL_USAGE_SUFFIX)
        return safe_format(template, self.template_args)

    @property
    def command(self):
        return self.decision.tool_input

    @property
    def output(self):
        return self.observation

    @property
    def template_args(self) -> Dict[str, str]:
        """Construct the dict used to render this output"""
        return {"command": self.command, "output": self.output}


class UseTerminal(Action):
    @classmethod
    def default(cls, llm: BaseLLM, prefix: Prefix, terminal: ZTerminal):
        return cls(
            name="Use the terminal (to run a command, not to edit a file)",
            output_type=TerminalOutput,
            chain=TerminalChain(
                llm=llm,
                prompt=TerminalPromptTemplate(prefix=prefix),
                terminal=terminal,
            ),
        )



"""Chain that interprets a prompt and executes bash code to perform bash operations."""
from typing import Dict, List

from langchain.llms.base import BaseLLM
from langchain.prompts import PromptTemplate
from pydantic import BaseModel, Extra

from ..prompts.chained import ChainedPromptTemplate
from .bash_action_prompt import BASH_COMMAND_TEMPLATE
from .general import ChoiceChain, ChoicePromptTemplate

POST_ACTION_PROMPT = """
{command}
{output}
```

You squint at the screen to determine whether:""".lstrip()

POST_ACTION_TEMPLATE = PromptTemplate(
    input_variables=["command", "output"],
    template=POST_ACTION_PROMPT,
)


COMMAND_RESULT_TEMPLATE = ChoicePromptTemplate(
    prefix=ChainedPromptTemplate("", BASH_COMMAND_TEMPLATE, POST_ACTION_TEMPLATE),
    choices=[
        "The command succeeded as intended",
        "The command failed",
    ],
)


BOOLS = ["True", "False"]


class ZOutputSuccessChain(ChoiceChain, BaseModel):
    """Checks whether or not output succeeded"""

    llm: BaseLLM
    """LLM wrapper to use."""
    prompt: ChoicePromptTemplate = COMMAND_RESULT_TEMPLATE

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    @property
    def output_keys(self) -> List[str]:
        """Output keys produced by this chain."""
        return ["visual_success"]

    def _return_dict(self, choice_num: int, choice: str) -> Dict[str, str]:
        return {"visual_success": BOOLS[choice_num - 1]}



"""Chain that interprets a prompt and executes bash code to perform bash operations."""
from typing import Dict, List

from langchain.chains.base import Chain
from langchain.llms.base import BaseLLM
from langchain.prompts.base import BasePromptTemplate
from pydantic import BaseModel, Extra

from .bash_action_prompt import SUCCESS_CRITERIA_TEMPLATE


class GetSuccessCriteriaChain(Chain, BaseModel):
    """Custom Bash chain for easier experimentation"""

    llm: BaseLLM
    """LLM wrapper to use."""
    prompt: BasePromptTemplate = SUCCESS_CRITERIA_TEMPLATE

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    @property
    def input_keys(self) -> List[str]:
        """Input keys to this chain."""
        return ["documentation", "task"]

    @property
    def output_keys(self) -> List[str]:
        """Output keys produced by this chain."""
        return ["checklist"]

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        output = self.llm(self.prompt.format(**inputs), stop=["\n\n"]).strip()

        return {"checklist": output}



from typing import Any, Dict, List

from langchain.chains.base import Chain
from langchain.chains.llm import LLMChain
from langchain.chains.sequential import SequentialChain
from langchain.llms.base import BaseLLM
from pydantic import BaseModel

from zamm.chains.general.boolean_switch import BooleanSwitchChain
from zamm.prompts.chained import ChainedPromptTemplate
from zamm.prompts.prefixed import Prefix

from .filesystem import FileSystemTool
from .prompt import NEW_CONTENTS_PROMPT, REPLACE_CONTENTS_PROMPT, WHICH_FILE_PROMPT


class AskFileChain(LLMChain, BaseModel):
    """Asks LLM for path of file to edit."""

    @property
    def input_keys(self) -> List[str]:
        """Input keys to this chain."""
        return self.prompt.input_variables

    @property
    def output_keys(self) -> List[str]:
        return ["file_path"]

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        file_path = (
            self.llm(self.prompt.format(**inputs), stop=["`", "\n"]).strip().strip("`")
        )
        return {"file_path": file_path}


class FileOutputChain(LLMChain, BaseModel):
    """Asks LLM for content of new file."""

    @property
    def input_keys(self) -> List[str]:
        """Input keys to this chain."""
        return self.prompt.input_variables + ["file_path"]

    @property
    def output_keys(self) -> List[str]:
        return self.prompt.input_variables + ["new_contents"]

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        new_contents = self.llm(self.prompt.format(**inputs), stop=["```"]).strip()
        FileSystemTool().write_file(inputs["file_path"], new_contents)
        return {**inputs, "new_contents": new_contents}


class ReadFileChain(Chain):
    @property
    def input_keys(self) -> List[str]:
        """Input keys this chain expects."""
        return ["file_path"]

    @property
    def output_keys(self) -> List[str]:
        """Output keys this chain expects."""
        return ["file_exists", "old_contents"]

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        fs = FileSystemTool()
        return fs.read_file(file_path=inputs["file_path"]).as_llm_output(
            contents_key="old_contents"
        )


class EditFileChain(BooleanSwitchChain):
    def condition_was_met(self, outputs: Dict[str, Any]) -> bool:
        """Condition is whether or not the specified file exists."""
        return outputs["file_exists"]

    @classmethod
    def default(cls, llm: BaseLLM, prefix: Prefix):
        ask_file = AskFileChain(
            llm=llm, prompt=ChainedPromptTemplate("", prefix, WHICH_FILE_PROMPT)
        )
        read_file = ReadFileChain()
        pick_and_read_file = SequentialChain(
            chains=[ask_file, read_file],
            input_variables=ask_file.input_keys,
            return_all=True,
        )
        edit_file = FileOutputChain(
            llm=llm, prompt=ChainedPromptTemplate("", prefix, REPLACE_CONTENTS_PROMPT)
        )
        new_file = FileOutputChain(
            llm=llm, prompt=ChainedPromptTemplate("", prefix, NEW_CONTENTS_PROMPT)
        )
        return cls(
            option_picker=pick_and_read_file,
            true_chain=edit_file,
            false_chain=new_file,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zamm',
 'zamm.actions',
 'zamm.actions.edit_file',
 'zamm.actions.note',
 'zamm.actions.use_terminal',
 'zamm.agents',
 'zamm.agents.tools',
 'zamm.chains',
 'zamm.chains.general',
 'zamm.chains.general.choice',
 'zamm.llms',
 'zamm.prompts',
 'zamm.resources',
 'zamm.resources.tutorials']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'langchain-visualizer>=0.0.12,<0.0.13',
 'openai>=0.26.4,<0.27.0',
 'pexpect>=4.8.0,<5.0.0',
 'pyyaml>=6.0,<7.0',
 'simple-term-menu>=1.6.1,<2.0.0',
 'tiktoken>=0.2.0,<0.3.0',
 'typer[all]>=0.7.0,<0.8.0',
 'ulid>=1.1,<2.0',
 'vcr-langchain>=0.0.15,<0.0.16']

entry_points = \
{'console_scripts': ['zamm = zamm.cli:app']}

setup_kwargs = {
    'name': 'zamm',
    'version': '0.0.3',
    'description': 'General automation driver',
    'long_description': '# ZAMM\n\nThis is an informal automation tool where you show GPT how to do something, and have it do it for you afterwards. This is good for boring but straightforward tasks that you haven\'t gotten around to writing a proper script to automate.\n\nWe are entering a time when our target audiences may include machines as well as humans. As such, this tool will generate tutorials that you can edit to make pleasant for both humans and LLMs alike to read.\n\n**This is an experimental tool, and has only been run on WSL Ubuntu so far.** It seems to work ok on the specific examples below. YMMV. Please feel free to add issues or PRs.\n\n## Quickstart\n\n`pipx` recommended over `pip` for install because it should allow you to run this with a different version of `langchain` than the one you might have installed:\n\n```bash\npipx install zamm\n```\n\nTeach GPT to do something:\n\n```bash\nzamm teach\n```\n\nYou will be roleplaying the LLM. The results of your interaction will be output as a Markdown tutorial file, which you can then edit to be more human-readable. See [this example](zamm/resources/tutorials/hello.md) of teaching the LLM how to create a "Hello world" script.\n\nAfterwards, you can tell the LLM to do a slightly different task using that same tutorial:\n\n```bash\nzamm execute --task \'Write a script goodbye.sh that prints out "Goodbye world". Execute it.\' --documentation zamm/resources/tutorials/hello.md\n```\n\nThis results in [this example transcript](demos/hello-transcript.md) of LLM interactions. **Note that GPT successfully generalizes from the tutorial to code in a completely different language based just on the difference in filenames.** Imagine having to manually add that feature to a script!\n\n### Using internal tutorials\n\nSelect any of the [prepackaged tutorials](zamm/resources/tutorials/) as documentation by prefacing their filename with `@internal`. The `.md` extension is optional.\n\nFor example:\n\n```bash\nzamm execute --task \'Protect the `main` branch\' --documentation @internal/branch-protection\n```\n\nto protect the `main` branch of the project in the current directory on Github. (Note that this tutorial was written in mind for ZAMM-built projects, so YMMV for using this on custom projects.)\n\n### Sessions\n\nSessions are recorded in case a crash happens, or if you want to change something up. On Linux, sessions are saved to `~/.local/share/zamm/sessions/`. To continue from the most recent session, run\n\n```bash\nzamm teach --last-session\n```\n\n### Free-styling\n\nYou can also simply tell the LLM to do something without teaching it to do so beforehand. However, this is a lot more brittle. An example of a free-style command that works:\n\n```bash\nzamm execute --task \'Write a script hello.py that prints out "Hello world". Execute it.\'\n```\n\nThe resulting transcript can be found [here](demos/freestyle-hello-transcript.md).\n\n## Prompting\n\nWhen a step is failing and you need faster iteration by repeatedly testing a single prompt, you can do so with the `prompt` command. First, write your prompt out to a file on disk. Then run this command:\n\n```bash\nzamm prompt --stop \'\\n\' --raw <path-to-prompt>\n```\n',
    'author': 'Amos Jun-yeung Ng',
    'author_email': 'me@amos.ng',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)



from typing import Any, Dict, List

from langchain.chains.base import Chain
from langchain.prompts.base import BasePromptTemplate
from langchain.schema import AgentFinish

from zamm.actions.base import Action
from zamm.agents.z_step import ZStepOutput
from zamm.prompts.dummy import DummyPromptTemplate

FINISH_LINE = "That's all! **Don't take any more steps** because the task is now done!"


class FinishChain(Chain):
    @property
    def input_keys(self) -> List[str]:
        """Input keys this chain expects."""
        return []

    @property
    def output_keys(self) -> List[str]:
        """Output keys this chain expects."""
        return []

    def _call(self, inputs: Dict[str, str]) -> Dict[str, str]:
        return {}


class FinishOutput(ZStepOutput):
    logger_template: BasePromptTemplate = DummyPromptTemplate()

    def log(self, **kwargs) -> str:
        return self.decision.log

    @property
    def template_args(self) -> Dict[str, str]:
        """Construct the dict used to render this output"""
        return {}

    @classmethod
    def from_chain_output(cls, output: Dict[str, Any]):
        """Construct the step from chain output"""
        return cls(
            decision=AgentFinish(
                return_values={"output": "Finished task."},
                log=FINISH_LINE,
            ),
            observation=None,
        )


class Finish(Action):
    @classmethod
    def default(cls):
        return cls(
            name="Declare the task done",
            output_type=FinishOutput,
            chain=FinishChain(),
        )



from typing import Any, Dict

from langchain.chains.llm import LLMChain
from langchain.llms.base import BaseLLM
from langchain.prompts.prompt import PromptTemplate

from zamm.chains.general import ChoiceChain, ChoicePromptTemplate, ErrorChain, NoOpChain
from zamm.chains.general.options_chain import OptionsChain

COMMAND_PROMPT_TEMPLATE = PromptTemplate(
    input_variables=["command"],
    template="The LLM would like to run the command `{command}`",
)

EDIT_COMMAND_PROMPT_TEMPLATE = PromptTemplate(
    input_variables=["command"],
    template="The LLM would like to run the command `{command}`\n\nReplace it with: ",
)

COMMAND_REVIEW_TEMPLATE = ChoicePromptTemplate(
    prefix=COMMAND_PROMPT_TEMPLATE,
    choice_prompt="",
    choices=["Proceed", "Edit command", "Stop (or Ctrl-C)"],
)


class HumanCommandReviewChain(OptionsChain):
    option_picker: ChoiceChain

    @property
    def choice_num_key(self) -> str:
        # todo: have this and action_chain inherit from the same base class
        return self.option_picker.choice_num_key

    def option_index(self, picker_output: Dict[str, Any]) -> int:
        """The index of the option chosen."""
        return int(picker_output[self.choice_num_key]) - 1

    def __init__(self, llm: BaseLLM, **kwargs):
        option_picker = ChoiceChain(
            llm=llm,
            prompt=COMMAND_REVIEW_TEMPLATE,
        )
        options = [
            NoOpChain(),
            LLMChain(
                llm=llm, output_key="command", prompt=EDIT_COMMAND_PROMPT_TEMPLATE
            ),
            ErrorChain(error_message="LLM danger detected."),
        ]
        super().__init__(option_picker=option_picker, options=options, **kwargs)



from typing import Any, List

from langchain.prompts import BasePromptTemplate, PromptTemplate

from zamm.prompts.prefixed import Prefix

from ..utils import f_join, safe_format


class ChainedPromptTemplate(BasePromptTemplate):
    joiner: str = ""
    subprompts: List[BasePromptTemplate]

    def __init__(self, joiner: str, *subprompts: Prefix):
        prompts: List[BasePromptTemplate] = []
        for subprompt in subprompts:
            if isinstance(subprompt, str):
                if subprompt != "":  # ignore empty strings
                    prompts.append(
                        PromptTemplate(input_variables=[], template=subprompt)
                    )
            elif isinstance(subprompt, BasePromptTemplate):
                prompts.append(subprompt)
            else:
                raise ValueError(f"Subprompt {subprompt} has unknown type")
        input_variables = list(
            set([var for subprompt in prompts for var in subprompt.input_variables])
        )
        super().__init__(
            joiner=joiner,
            subprompts=prompts,
            input_variables=input_variables,
        )

    def format(self, **kwargs: Any) -> str:
        """Format the prompt with the inputs."""
        results = [safe_format(prompt, kwargs) for prompt in self.subprompts]
        return f_join(self.joiner, results)

    @property
    def _prompt_type(self) -> str:
        """Return the prompt type key."""
        return "chained"



from typing import Any, Dict, List, Optional

from langchain.agents.agent import AgentExecutor
from langchain.agents.tools import Tool
from langchain.llms.base import BaseLLM
from langchain.schema import AgentAction, AgentFinish

from zamm.actions.base import Action
from zamm.actions.edit_file import EditFile
from zamm.actions.finish import Finish
from zamm.actions.note import MakeNote
from zamm.actions.use_terminal import UseTerminal, ZTerminal
from zamm.chains.bash_action_prompt import (
    EMPLOYEE_TEACHING_INTRO_TEMPLATE,
    FOLLOW_INSTRUCTIONS_TEMPLATE,
)
from zamm.chains.general import ActionChain
from zamm.chains.general.choice.base import ChoiceChain
from zamm.chains.general.choice.prompt import ChoicePromptTemplate
from zamm.prompts.chained import ChainedPromptTemplate
from zamm.prompts.prefixed import Prefix
from zamm.utils import f_join

from .base import CustomAgent
from .memory import AgentMemory, BaseAgentMemory
from .step import StepOutput


class ZammEmployeeBrain(CustomAgent):
    @property
    def _agent_type(self) -> str:
        """Return Identifier of agent type."""
        return "zamm-employee"

    def _get_next_action(self, full_inputs: Dict[str, str]) -> AgentAction:
        raise NotImplementedError()

    def _construct_scratchpad_structured(self, memory: BaseAgentMemory) -> str:
        logs = []
        steps = memory.steps()
        for i, output in enumerate(steps):
            previous = steps[i - 1] if i > 0 else None
            next = steps[i + 1] if i < len(steps) - 1 else None
            logs.append(output.log(previous=previous, next=next))
        scratch = f_join("\n", logs)
        if scratch == "":
            return scratch
        return "\n" + scratch + "\n"


def default_action_chain(
    llm: BaseLLM,
    prefix: Prefix,
    terminal: ZTerminal,
    choice_prompt: str = "You now contemplate your next step:",
):
    actions: List[Action] = [
        MakeNote.default(llm=llm, prefix=prefix),
        UseTerminal.default(llm=llm, prefix=prefix, terminal=terminal),
        EditFile.default(llm=llm, prefix=prefix),
        Finish.default(),
    ]

    action_choice_template = ChoicePromptTemplate(
        prefix=ChainedPromptTemplate("", prefix, choice_prompt),
        choices=[action.choice_text for action in actions],
    )
    action_choice = ChoiceChain(
        llm=llm,
        prompt=action_choice_template,
        choice_num_key="action_num",
        choice_key="action",
    )

    return ActionChain(option_picker=action_choice, actions=actions)


class ZammEmployee(AgentExecutor):
    agent: ZammEmployeeBrain
    terminal: ZTerminal
    max_iterations: Optional[int] = 100

    def __init__(
        self,
        llm: BaseLLM,
        tools: Optional[List[Tool]] = None,
        terminal_safe_mode: bool = True,
        **kwargs,
    ):
        tools = tools if tools else []
        super().__init__(
            agent=ZammEmployeeBrain.from_llm(llm),
            tools=tools,
            terminal=ZTerminal(safe_mode=terminal_safe_mode),
            **kwargs,
        )

    @property
    def input_keys(self) -> List[str]:
        return ["task", "documentation"]

    def _custom_take_next_step(self, memory: BaseAgentMemory) -> StepOutput:
        if "documentation" in memory.inputs:
            prefix = FOLLOW_INSTRUCTIONS_TEMPLATE
        else:
            prefix = EMPLOYEE_TEACHING_INTRO_TEMPLATE

        action_chain = default_action_chain(
            llm=self.agent.llm_chain.llm,
            prefix=prefix,
            terminal=self.terminal,
        )
        results = action_chain(
            {
                **memory.inputs,
                "agent_scratchpad": self.agent._construct_scratchpad_structured(memory),
            }
        )
        return results[action_chain.step_output_key]

    def _return_structured(
        self, output: StepOutput, memory: AgentMemory
    ) -> Dict[str, Any]:
        memory.add_step(output)
        task = memory.inputs["task"]
        logs = self.agent._construct_scratchpad_structured(memory)
        tutorial = f"""
Say you want to do the following task:

> {task}

You can do so by following these steps:

{logs}
""".strip()
        return {"output": tutorial}

    def _validate_inputs(self, inputs: Dict[str, str]) -> None:
        """Skip validation because some inputs are optional."""

    def _call(self, inputs: Dict[str, str]) -> Dict[str, Any]:
        """Stripped down version of AgentExecutor._call."""
        # Do any preparation necessary when receiving a new input.
        self.agent.prepare_for_new_call()

        memory = AgentMemory(inputs=inputs)
        iterations = 0
        # We now enter the agent loop (until it returns something).
        while self._should_continue(iterations):
            step_result = self._custom_take_next_step(memory)
            # If the tool chosen is the finishing tool, then we end and return.
            if isinstance(step_result.decision, AgentFinish):
                return self._return_structured(step_result, memory)

            memory.add_step(step_result)
            iterations += 1
        output = self.agent.return_stopped_response(
            self.early_stopping_method, memory.as_intermediate_steps(), **inputs
        )
        return self._return(output, memory.as_intermediate_steps())



from typing import Any, List

from fvalues import F

from zamm.prompts import PrefixedPromptTemplate
from zamm.utils import f_join


class ChoicePrompt(F):
    prefix_text: str
    choices: List[str]

    def __new__(cls, f_str: F, prefix_text: str, choices: List[str]):
        result = super().__new__(cls, f_str, parts=f_str.parts)
        result.prefix_text = prefix_text
        result.choices = choices
        return result


class ChoicePromptTemplate(PrefixedPromptTemplate):
    choices: List[str]
    choice_prompt: str = "You decide it's #"

    @property
    def _prompt_type(self) -> str:
        """Return the prompt type key."""
        return "choice"

    @property
    def choices_text(self) -> str:
        return "\n".join([f"{i+1}. {choice}" for i, choice in enumerate(self.choices)])

    def format(self, **kwargs: Any) -> str:
        prefix_text = self.prefix.format(**kwargs)
        prompt = f_join(
            "\n\n",
            [
                prefix_text,
                self.choices_text,
                self.choice_prompt,
            ],
        )
        return ChoicePrompt(prompt, prefix_text, self.choices)



"""Rich structured data about each iteration of the agent loop."""

from typing import Any, Optional, Tuple, Union

from langchain.schema import AgentAction, AgentFinish
from pydantic import BaseModel


class StepOutput(BaseModel):
    """The output produced by a single step of the agent loop.

    You can subclass this to store rich structured data instead of relying solely on
    AgentAction's str inputs.
    """

    decision: Union[AgentAction, AgentFinish]
    """The decision taken during the step.

    This could be deciding to use a tool, or deciding to end the loop.
    """
    observation: Optional[str]
    """The observed effects of the previous action.

    Only present if the decision taken wasn't to end the loop.
    """

    @property
    def is_finish(self) -> bool:
        """Whether this step signals the end of the agent loop."""
        return isinstance(self.decision, AgentFinish)

    def log(self, **kwargs: Any) -> str:
        """Produce a log for this step.

        This is usually useful for constructing the scratchpad. The kwargs are there to
        allow for flexibility during scratchpad construction. Subclass `StepOutput` and
        override this function to make use of them.
        """
        if len(kwargs) > 0:
            raise ValueError("The default log doesn't take any options")
        assert self.observation is not None
        # this is just an example because this function isn't actually used in any of
        # the current Agent logic, but if kwargs were to contain `observation_prefix`
        # and `llm_prefix`, then we could recreate the scratchpad logic in
        # `Agent._construct_scratchpad`
        return self.decision.log + "\n" + self.observation

    def as_intermediate_step(self) -> Tuple[AgentAction, str]:
        """Compatibility function for existing action-observation tuple."""
        assert isinstance(self.decision, AgentAction)
        assert self.observation is not None  # should follow from first assert
        return (self.decision, self.observation)
